import sys
import os
from math import pi

from PyQt5.QtCore import Qt, QEventLoop
from PyQt5.QtWidgets import QApplication

from vortex import Range, get_console_logger as gcl
from vortex.scan import RasterScanConfig, RasterScan
from vortex.engine import EngineConfig, Engine, SpectraHDF5StackEndpoint, AscanHDF5StackEndpoint, SpectraStackEndpoint, AscanStackEndpoint, NullEndpoint
from vortex.storage import HDF5StackUInt16, HDF5StackInt8, HDF5StackConfig, HDF5StackHeader, SimpleStackUInt16, SimpleStackInt8, SimpleStackConfig, SimpleStackHeader
from vortex.format import FormatPlanner, FormatPlannerConfig, StackFormatExecutorConfig, StackFormatExecutor, SimpleSlice

# hack to simplify running demos
sys.path.append(os.path.dirname(__file__))

from _common.engine import setup_logging, StandardEngineParams, DEFAULT_ENGINE_PARAMS, BaseEngine

class OCTEngine(BaseEngine):
    def __init__(self, cfg: StandardEngineParams, args):
        super().__init__(cfg)

        #
        # scan
        #

        raster_sc = RasterScanConfig()
        raster_sc.bscans_per_volume = cfg.bscans_per_volume
        raster_sc.ascans_per_bscan = cfg.ascans_per_bscan
        raster_sc.bscan_extent = Range(-cfg.scan_dimension, cfg.scan_dimension)
        raster_sc.volume_extent = Range(-cfg.scan_dimension, cfg.scan_dimension)
        raster_sc.bidirectional_segments = cfg.bidirectional
        raster_sc.bidirectional_volumes = cfg.bidirectional
        raster_sc.samples_per_second = cfg.swept_source.triggers_per_second
        raster_sc.loop = False

        raster_scan = RasterScan()
        raster_scan.initialize(raster_sc)
        self._raster_scan = raster_scan

        #
        # output setup
        #

        # format planners
        fc = FormatPlannerConfig()
        fc.segments_per_volume = cfg.bscans_per_volume
        fc.records_per_segment = cfg.ascans_per_bscan
        fc.adapt_shape = False

        stack_format = FormatPlanner(gcl('format', cfg.log_level))
        stack_format.initialize(fc)

        # format executors
        cfec = StackFormatExecutorConfig()
        cfe = StackFormatExecutor()
        cfe.initialize(cfec)

        self._stack_spectra_storage = None
        self._stack_ascan_storage = None
        endpoints = []

        if args.format in ['matlab', 'hdf5']:

            fmt = HDF5StackHeader.MATLAB if args.format == 'matlab' else HDF5StackHeader.Empty
            suffix = '.mat' if args.format == 'matlab' else '.h5'

            if not args.no_save_spectra:
                # save spectra data
                h5sc = HDF5StackConfig()
                h5sc.shape = (cfg.bscans_per_volume, cfg.ascans_per_bscan, self._process.config.samples_per_record, 1)
                h5sc.header = fmt
                h5sc.path = f'{args.prefix}spectra{suffix}'

                self._stack_spectra_storage = HDF5StackUInt16(gcl('hdf5-spectra', cfg.log_level))
                self._stack_spectra_storage.open(h5sc)
                endpoints.append(SpectraHDF5StackEndpoint(cfe, self._stack_spectra_storage, log=gcl('hdf5-spectra', cfg.log_level)))

            if not args.no_save_ascans:
                # save ascan data
                h5sc = HDF5StackConfig()
                h5sc.shape = (cfg.bscans_per_volume, cfg.ascans_per_bscan, self._process.config.samples_per_ascan, 1)
                h5sc.header = fmt
                h5sc.path = f'{args.prefix}ascans{suffix}'

                self._stack_ascan_storage = HDF5StackInt8(gcl('hdf5-ascan', cfg.log_level))
                self._stack_ascan_storage.open(h5sc)
                endpoints.append(AscanHDF5StackEndpoint(cfe, self._stack_ascan_storage, log=gcl('hdf5-ascan', cfg.log_level)))

        else:


            if not args.no_save_spectra:
                # save spectra data
                npsc = SimpleStackConfig()
                npsc.shape = (cfg.bscans_per_volume, cfg.ascans_per_bscan, self._process.config.samples_per_record)
                npsc.header = SimpleStackHeader.NumPy
                npsc.path = f'{args.prefix}spectra.npy'

                self._stack_spectra_storage = SimpleStackUInt16(gcl('npy-spectra', cfg.log_level))
                self._stack_spectra_storage.open(npsc)
                endpoints.append(SpectraStackEndpoint(cfe, self._stack_spectra_storage, log=gcl('npy-spectra', cfg.log_level)))

            if not args.no_save_ascans:
                # save ascan data
                npsc = SimpleStackConfig()
                npsc.shape = (cfg.bscans_per_volume, cfg.ascans_per_bscan, self._process.config.samples_per_ascan)
                npsc.header = SimpleStackHeader.NumPy
                npsc.path = f'{args.prefix}ascans.npy'

                self._stack_ascan_storage = SimpleStackInt8(gcl('npy-ascan', cfg.log_level))
                self._stack_ascan_storage.open(npsc)
                endpoints.append(AscanStackEndpoint(cfe, self._stack_ascan_storage, log=gcl('npy-ascan', cfg.log_level)))

        #
        # engine setup
        #

        ec = EngineConfig()
        ec.add_acquisition(self._acquire, [self._process])
        ec.add_processor(self._process, [stack_format])
        ec.add_formatter(stack_format, endpoints or [NullEndpoint()])
        ec.add_io(self._io_out, lead_samples=round(cfg.galvo_delay * self._io_out.config.samples_per_second))
        ec.add_io(self._strobe)

        ec.preload_count = cfg.preload_count
        ec.records_per_block = cfg.ascans_per_block
        ec.blocks_to_allocate = cfg.blocks_to_allocate
        ec.blocks_to_acquire = cfg.blocks_to_acquire

        ec.galvo_output_channels = len(self._io_out.config.channels)
        ec.galvo_input_channels = 0

        engine = Engine(gcl('engine', cfg.log_level))
        self._engine = engine

        engine.initialize(ec)
        engine.prepare()

    def run(self):

        self._engine.scan_queue.append(self._raster_scan)

        self._engine.start()
        self._engine.wait()
        self._engine.stop()

        if self._stack_spectra_storage:
            self._stack_spectra_storage.close()
        if self._stack_ascan_storage:
            self._stack_ascan_storage.close()

if __name__ == '__main__':
    setup_logging()

    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
    parser = ArgumentParser(description='save volume to disk', formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('--format', choices=['matlab', 'hdf5', 'numpy'], default='numpy', help='file format for saving')
    parser.add_argument('--no-save-ascans', action='store_true', help='Do not save A-scans')
    parser.add_argument('--no-save-spectra', action='store_true', help='Do not save spectra')
    parser.add_argument('--prefix', default='', help='prefix for output file names')
    args = parser.parse_args()

    engine = OCTEngine(DEFAULT_ENGINE_PARAMS, args)
    engine.run()
