/** \rst

    CPU-based OCT processor

    This OCT processor uses xtensor and FFTW to perform OCT processing without
    CUDA support.  It is reasonably slow, not optimized, and is outdated. It is
    intended for debugging purposes only or for use if non-CUDA OCT processing
    must be done (e.g., on a laptop without CUDA capability).

    TODO: Update to the new block structure.

 \endrst */

 #pragma once

#include <memory>
#include <thread>
#include <future>
#include <shared_mutex>
#include <functional>

#include <tbb/parallel_for.h>

#include <xtensor/xindex_view.hpp>
#include <xtensor/xcomplex.hpp>

#include <vortex/process/base.hpp>

#include <vortex/driver/fftw.hpp>

#include <vortex/util/cast.hpp>
#include <vortex/util/sync.hpp>

namespace vortex::process {

    template<typename config_t_, typename input_assignment_t_, typename output_assignment_t_>
    class cpu_processor_t : processor_t<config_t_> {
    public:

        using config_t = config_t_;
        using input_assignment_t = input_assignment_t_;
        using output_assignment_t = output_assignment_t_;

        using input_element_t = typename input_assignment_t::element_t;
        using output_element_t = typename output_assignment_t::element_t;

        static_assert(std::is_same_v<
            typename input_assignment_t::block_t,
            typename output_assignment_t::block_t>,
            "input and output assignments must have the same block type");

        using block_t = typename input_assignment_t::block_t;
        using callback_t = std::function<void(block_t*)>;

    protected:
        struct job_t {
            block_t* block;
            callback_t callback;
        };

        struct slot_t {
            xt::xtensor<output_element_t, 2> mag;
            xt::xtensor<std::complex<output_element_t>, 2> pre_fft, post_fft;

            //event_t* pre_average = nullptr;
            //event_t* post_average = nullptr;
            //
            //void wait_for_prior_average() {
            //    if (pre_average) {
            //        pre_average->wait();
            //    }
            //}
            //
            //void notify_average_is_complete() {
            //    if (post_average) {
            //        post_average->set();
            //    }
            //}
            //
            //void rearm() {
            //    if (post_average) {
            //        post_average->unset();
            //    }
            //}

        };
    public:

        virtual ~cpu_processor_t() {
            _shutdown_workers();
        }

        virtual void initialize(config_t config) {
            config.validate();

            _config = std::move(config);

            _recalculate();
        }

        virtual void change(config_t& new_config) {
            new_config.validate();

            _shutdown_workers();
            std::unique_lock<std::shared_mutex> lock(_config_mutex);

            _recalculate();
        }

        virtual void next(block_t& block) {
            std::shared_lock<std::shared_mutex> lock(_config_mutex);
            std::unique_lock<std::mutex> lock2(_slot_mutex);

            _process_block(_slot, block, true);

            //// wait for a processing slot
            //auto slot = _slots.pop();
            //slot->rearm();
            //
            //// process the block
            //_process_block(*slot, block);
            //
            //// return the slot
            //_slots.push(slot);
        }

        virtual void next_async(block_t* block) {
            next_async(block, callback_t{});
        }
        virtual void next_async(block_t* block, sync::event_t* event) {
            event->unset();
            next_async(block, [this, event](block_t*) { event->set(); });
        }
        virtual void next_async(block_t* block, callback_t&& callback) {
            // prevent the config from changing while jobs are active
            std::shared_lock<std::shared_mutex> lock(_config_mutex);

            _jobs.push({ block, std::forward<callback_t>(callback) });
            if (!_worker.joinable()) {
                // launch background thread
                _worker_ready.unset();
                _worker = std::thread([&] { _worker_loop(); });

                // wait for worker to acquire a shared config lock
                _worker_ready.wait();
            }
        }

    protected:

        void _recalculate() {
            // set up A-scan averaging for constant frequency artifact removal
            if (_config.average_ascan_window > 0) {
                if (_average_ascan.shape(0) != _config.average_ascan_window) {
                    // TODO: adjust buffer while retaining history
                    _average_ascan.resize({ _config.average_ascan_window, _config.samples_per_ascan() });
                    _average_ascan_index = _average_ascan_count = 0;
                }
            } else {
                // release buffer
                _average_ascan = {};
                _average_ascan_index = _average_ascan_count = 0;
            }

            // set up slots
            //_slots.clear();
            //_slot_buffer.resize(_config.slots);
            //_slot_events.resize(_config.slots);
            //for (ptrdiff_t i = 0; i < _slot_buffer.size(); i++) {
            //    auto& slot = _slot_buffer[i];
                auto& slot = _slot;

                // allocate intermediate buffers
                slot.mag.resize(_config.shape);
                slot.pre_fft.resize(_config.shape);
                slot.post_fft.resize(_config.shape);

            //    // initialize events
            //    if (_config.slots > 1) {
            //        slot.pre_average = &_slot_events[(i - 1) % _config.slots];
            //        slot.post_average = &_slot_events[i];
            //    }

            //    // make available
            //    _slots.push(*slot);
            //}

            // resampling
            if (_config.resampling_samples.size() > 0) {
                if (_config.samples_per_ascan() < 2) {
                    throw std::runtime_error(fmt::format("at least two samples per ascan required for resampling: {}", _config.samples_per_ascan()));
                }

                // the sample index before the requested index
                _resample_before_index = xt::cast<size_t>(xt::floor(_config.resampling_samples));
                // the sample index after the requested index
                _resample_after_index = _resample_before_index + 1;

                // compute the sampling ratio
                _resample_after_weight = _config.resampling_samples - _resample_before_index;
                _resample_before_weight = (1 - _resample_after_weight);

                // fix the edge cases
                for (size_t i = 0; i < _config.samples_per_ascan(); i++) {
                    if (_resample_before_index(i) < 0) {
                        // clamp to first sample
                        _resample_before_index(i) = _resample_after_index(i) = 0;
                    } else if (_resample_after_index(i) >= downcast<ptrdiff_t>(_config.samples_per_ascan())) {
                        // clamp to last sample
                        _resample_before_index(i) = _resample_after_index(i) = std::max(0LL, downcast<ptrdiff_t>(_config.samples_per_ascan()) - 1);
                    }
                }
            } else {
                // release
                _resample_before_index = {};
                _resample_after_index = {};
                _resample_before_weight = {};
                _resample_after_weight = {};
            }

            // FFT planning
            // NOTE: FFTW requires fully allocated buffers for plan generation
            // TODO: move plan into slot and switch to inplace execution (let it allocate its own internal buffers)
            xt::xtensor<std::complex<output_element_t>, 2> a, b;
            a.resize(_config.shape);
            b.resize(_config.shape);

            auto n = downcast<int>(_config.samples_per_ascan());
            _ifft.inverse(
                downcast<int>(_config.ascans_per_block()),   // number of FFTs
                { 1, 1 },                            // stride between successive samples
                { n, n },                            // stride between successive A-scans
                { n },                               // length of FFT
                a.data(), b.data()                   // data pointers
            );
        }

        void _shutdown_workers() {
            _jobs.clear();
            _jobs.finish();

            if (_worker.joinable()) {
                _worker.join();
            }
        }

        virtual void _process_block(slot_t& slot, block_t& block, bool append_history) {
            // determine if this block length is acceptable
            if (block.length > _config.ascans_per_block()) {
                throw std::runtime_error(fmt::format("block is larger than maximum configured size: {} > {}", block.length, _config.ascans_per_block()));
            }
            auto rows = std::min(block.length, _config.ascans_per_block());

            // setup views into buffers
            auto& mag = xt::view(slot.mag, xt::range(0, rows), xt::all());
            auto& pre_fft = xt::view(slot.pre_fft, xt::range(0, rows), xt::all());
            auto& post_fft = xt::view(slot.post_fft, xt::range(0, rows), xt::all());
            auto& output = output_assignment_t::select(block).data();

            // copy in the initial values
            // TODO: avoid this initial copy
            mag = input_assignment_t::select(block).data();

            // resample
            if (_config.resampling_samples.size() > 0) {
                // resample each column
                tbb::parallel_for(0LL, downcast<ptrdiff_t>(rows), [&](ptrdiff_t r) {
                    auto before = xt::index_view(xt_row(mag, r), _resample_before_index);
                    auto after = xt::index_view(xt_row(mag, r), _resample_after_index);
                    xt_row(mag, r) = _resample_before_weight * before + _resample_after_weight * after;
                });
            }

            // average A-scan
            //slot.wait_for_prior_average();
            if (_average_ascan.size() > 0) {
                std::unique_lock<std::mutex> lock2(_average_ascan_mutex);

                if (append_history) {
                    // update the history window
                    size_t count = std::min(rows, _average_ascan.shape(0));
                    size_t released = 0;

                    while (released < count) {
                        // never copy off the end of the window buffer
                        size_t available = std::min(_average_ascan.shape(0) - _average_ascan_index, count - released);
                        // start from the end of the mag for recency
                        size_t index = rows - count + released;

                        // perform the copy
                        auto dst = xt::view(_average_ascan, xt::range(_average_ascan_index, _average_ascan_index + available), xt::all());
                        auto src = xt::view(mag, xt::range(index, index + available), xt::all());
                        dst = src;

                        _average_ascan_index = (_average_ascan_index + available) % _average_ascan.shape(0);
                        released += available;
                    }
                    _average_ascan_count = std::min(_average_ascan_count + count, _average_ascan.shape(0));
                }

                // remove the average from this block
                if (_average_ascan_count > 0) {
                    auto avg = xt::mean(xt::view(_average_ascan, xt::range(0, _average_ascan_count), xt::all()), 0, xt::evaluation_strategy::immediate);
                    mag -= xt::view(avg, xt::newaxis(), xt::all());
                }
            }
            //slot.notify_average_is_complete();

            // spectral filtering
            if (_config.spectral_filter.size() > 0) {
                pre_fft = mag * xt::view(_config.spectral_filter, xt::newaxis(), xt::all());
            } else {
                pre_fft = mag;
            }

            // inverse FFT without normalization
            // NOTE: FFT operates on the slot buffers because the number of planned FFTs is fixed
            _ifft.execute(slot.pre_fft.data(), slot.post_fft.data());

            // normalize + abs + log10
            output = 20 * xt::log10(xt::abs(post_fft) / downcast<output_element_t>(_config.samples_per_ascan()));
        }

        void _worker_loop() {
            std::shared_lock<std::shared_mutex> lock(_config_mutex);
            _worker_ready.set();

            job_t job;
            while (_jobs.pop(job)) {
                std::unique_lock<std::mutex> lock2(_slot_mutex);

                _process_block(_slot, *job.block, true);

                if (job.callback) {
                    job.callback(job.block);
                }
            }
        }

        sync::queue_t<job_t> _jobs;

        std::thread _worker;
        sync::event_t _worker_ready;

        std::mutex _slot_mutex;
        slot_t _slot;

        //sync::queue<slot_t*> _slots;
        //std::vector<event_t> _slot_events;
        //std::vector<slot_t> _slots;

        // resampling
        xt::xtensor<output_element_t, 1> _resample_before_weight, _resample_after_weight;
        xt::xtensor<ptrdiff_t, 1> _resample_before_index, _resample_after_index;

        // averaging
        std::mutex _average_ascan_mutex;
        xt::xtensor<output_element_t, 2> _average_ascan;
        size_t _average_ascan_index, _average_ascan_count;

        // FFT
        fft::fftw_plan_t<output_element_t> _ifft;

        std::shared_mutex _config_mutex;

    };


}